"""
Tests for logical constraint operators.

Tests for the AndConstraint, OrConstraint, and NotConstraint classes,
as well as the __and__, __or__, and __invert__ operators on ConstraintABC.
Tests cover both operator creation and constraint evaluation logic.
"""

from typing import cast

import numpy as np
from astropy.coordinates import SkyCoord  # type: ignore[import-untyped]
from astropy.time import Time  # type: ignore[import-untyped]

from across.tools.core.enums import ConstraintType
from across.tools.ephemeris import Ephemeris
from across.tools.visibility.constraints import (
    AndConstraint,
    NotConstraint,
    OrConstraint,
    XorConstraint,
)
from across.tools.visibility.constraints.base import ConstraintABC


class TestAndOperator:
    """Test suite for the & operator creating AndConstraint."""

    def test_and_operator_creates_and_constraint(self, logical_and_constraint: AndConstraint) -> None:
        """Test that & operator creates an AndConstraint."""
        assert isinstance(logical_and_constraint, AndConstraint)

    def test_and_operator_with_two_constraints_count(self, logical_and_constraint: AndConstraint) -> None:
        """Test & operator creates AndConstraint with two constraints."""
        assert len(logical_and_constraint.constraints) == 2

    def test_and_operator_with_two_constraints_first(
        self,
        logical_and_constraint: AndConstraint,
        logical_sun_constraint: ConstraintABC,
    ) -> None:
        """Test & operator keeps sun as first constraint."""
        assert logical_and_constraint.constraints[0] is logical_sun_constraint

    def test_and_operator_with_two_constraints_second(
        self,
        logical_and_constraint: AndConstraint,
        logical_moon_constraint: ConstraintABC,
    ) -> None:
        """Test & operator keeps moon as second constraint."""
        assert logical_and_constraint.constraints[1] is logical_moon_constraint


class TestOrOperator:
    """Test suite for the | operator creating OrConstraint."""

    def test_or_operator_creates_or_constraint(self, logical_or_constraint: OrConstraint) -> None:
        """Test that | operator creates an OrConstraint."""
        assert isinstance(logical_or_constraint, OrConstraint)

    def test_or_operator_with_two_constraints_count(self, logical_or_constraint: OrConstraint) -> None:
        """Test | operator creates OrConstraint with two constraints."""
        assert len(logical_or_constraint.constraints) == 2

    def test_or_operator_with_two_constraints_first(
        self,
        logical_or_constraint: OrConstraint,
        logical_sun_constraint: ConstraintABC,
    ) -> None:
        """Test | operator keeps sun as first constraint."""
        assert logical_or_constraint.constraints[0] is logical_sun_constraint

    def test_or_operator_with_two_constraints_second(
        self,
        logical_or_constraint: OrConstraint,
        logical_moon_constraint: ConstraintABC,
    ) -> None:
        """Test | operator keeps moon as second constraint."""
        assert logical_or_constraint.constraints[1] is logical_moon_constraint


class TestNotOperator:
    """Test suite for the ~ operator creating NotConstraint."""

    def test_not_operator_creates_not_constraint(self, logical_not_constraint: NotConstraint) -> None:
        """Test that ~ operator creates a NotConstraint."""
        assert isinstance(logical_not_constraint, NotConstraint)

    def test_not_operator_with_constraint(
        self,
        logical_not_constraint: NotConstraint,
        logical_sun_constraint: ConstraintABC,
    ) -> None:
        """Test ~ operator creates NotConstraint with correct constraint."""
        assert logical_not_constraint.constraint is logical_sun_constraint


class TestXorOperator:
    """Test suite for the ^ operator creating XorConstraint."""

    def test_xor_operator_creates_xor_constraint(self, logical_xor_constraint: XorConstraint) -> None:
        """Test that ^ operator creates an XorConstraint."""
        assert isinstance(logical_xor_constraint, XorConstraint)

    def test_xor_operator_with_two_constraints_count(self, logical_xor_constraint: XorConstraint) -> None:
        """Test ^ operator creates XorConstraint with two constraints."""
        assert len(logical_xor_constraint.constraints) == 2

    def test_xor_operator_with_two_constraints_first(
        self,
        logical_xor_constraint: XorConstraint,
        logical_sun_constraint: ConstraintABC,
    ) -> None:
        """Test ^ operator keeps sun as first constraint."""
        assert logical_xor_constraint.constraints[0] is logical_sun_constraint

    def test_xor_operator_with_two_constraints_second(
        self,
        logical_xor_constraint: XorConstraint,
        logical_moon_constraint: ConstraintABC,
    ) -> None:
        """Test ^ operator keeps moon as second constraint."""
        assert logical_xor_constraint.constraints[1] is logical_moon_constraint


class TestConstraintTypeValues:
    """Test suite for constraint type enum values."""

    def test_constraint_type_for_and(self, logical_and_constraint: AndConstraint) -> None:
        """Test that AND constraint has correct ConstraintType."""
        assert logical_and_constraint.name == ConstraintType.AND

    def test_constraint_type_for_or(self, logical_or_constraint: OrConstraint) -> None:
        """Test that OR constraint has correct ConstraintType."""
        assert logical_or_constraint.name == ConstraintType.OR

    def test_constraint_type_for_not(self, logical_not_constraint: NotConstraint) -> None:
        """Test that NOT constraint has correct ConstraintType."""
        assert logical_not_constraint.name == ConstraintType.NOT

    def test_constraint_type_for_xor(self, logical_xor_constraint: XorConstraint) -> None:
        """Test that XOR constraint has correct ConstraintType."""
        assert logical_xor_constraint.name == ConstraintType.XOR


class TestAndConstraint:
    """Test suite for AndConstraint functionality."""

    def test_and_constraint_empty_constraints(
        self, mock_ephemeris: Ephemeris, time_array: Time, sky_coord: SkyCoord
    ) -> None:
        """Test AndConstraint with empty constraints list."""
        and_constraint = AndConstraint(constraints=[])
        result = and_constraint(time=time_array, ephemeris=mock_ephemeris, coordinate=sky_coord)
        assert np.all(~result)  # All False since empty

    def test_and_constraint_single_constraint(
        self,
        mock_ephemeris: Ephemeris,
        time_array: Time,
        sky_coord: SkyCoord,
        true_constraint: ConstraintABC,
    ) -> None:
        """Test AndConstraint with a single constraint returns same result."""
        and_constraint = AndConstraint(constraints=[true_constraint])
        result = and_constraint(time=time_array, ephemeris=mock_ephemeris, coordinate=sky_coord)
        expected = true_constraint(time=time_array, ephemeris=mock_ephemeris, coordinate=sky_coord)
        assert np.array_equal(result, expected)

    def test_and_constraint_logic_true_and_true(
        self,
        mock_ephemeris: Ephemeris,
        time_array: Time,
        sky_coord: SkyCoord,
        true_constraint: ConstraintABC,
    ) -> None:
        """Test AndConstraint logic for True AND True."""
        and_constraint = AndConstraint(constraints=[true_constraint, true_constraint])
        result = and_constraint(time=time_array, ephemeris=mock_ephemeris, coordinate=sky_coord)
        assert np.all(result)

    def test_and_constraint_logic_true_and_false(
        self,
        mock_ephemeris: Ephemeris,
        time_array: Time,
        sky_coord: SkyCoord,
        true_constraint: ConstraintABC,
        false_constraint: ConstraintABC,
    ) -> None:
        """Test AndConstraint logic for True AND False."""
        and_constraint = AndConstraint(constraints=[true_constraint, false_constraint])
        result = and_constraint(time=time_array, ephemeris=mock_ephemeris, coordinate=sky_coord)
        assert np.all(~result)

    def test_and_constraint_logic_false_and_false(
        self,
        mock_ephemeris: Ephemeris,
        time_array: Time,
        sky_coord: SkyCoord,
        false_constraint: ConstraintABC,
    ) -> None:
        """Test AndConstraint logic for False AND False."""
        and_constraint = AndConstraint(constraints=[false_constraint, false_constraint])
        result = and_constraint(time=time_array, ephemeris=mock_ephemeris, coordinate=sky_coord)
        assert np.all(~result)


class TestOrConstraint:
    """Test suite for OrConstraint functionality."""

    def test_or_constraint_empty_constraints(
        self, mock_ephemeris: Ephemeris, time_array: Time, sky_coord: SkyCoord
    ) -> None:
        """Test OrConstraint with empty constraints list."""
        or_constraint = OrConstraint(constraints=[])
        result = or_constraint(time=time_array, ephemeris=mock_ephemeris, coordinate=sky_coord)
        assert np.all(~result)  # All False since empty

    def test_or_constraint_single_constraint(
        self,
        mock_ephemeris: Ephemeris,
        time_array: Time,
        sky_coord: SkyCoord,
        true_constraint: ConstraintABC,
    ) -> None:
        """Test OrConstraint with a single constraint returns same result."""
        or_constraint = OrConstraint(constraints=[true_constraint])
        result = or_constraint(time=time_array, ephemeris=mock_ephemeris, coordinate=sky_coord)
        expected = true_constraint(time=time_array, ephemeris=mock_ephemeris, coordinate=sky_coord)
        assert np.array_equal(result, expected)

    def test_or_constraint_logic_true_or_true(
        self,
        mock_ephemeris: Ephemeris,
        time_array: Time,
        sky_coord: SkyCoord,
        true_constraint: ConstraintABC,
    ) -> None:
        """Test OrConstraint logic for True OR True."""
        or_constraint = OrConstraint(constraints=[true_constraint, true_constraint])
        result = or_constraint(time=time_array, ephemeris=mock_ephemeris, coordinate=sky_coord)
        assert np.all(result)

    def test_or_constraint_logic_true_or_false(
        self,
        mock_ephemeris: Ephemeris,
        time_array: Time,
        sky_coord: SkyCoord,
        true_constraint: ConstraintABC,
        false_constraint: ConstraintABC,
    ) -> None:
        """Test OrConstraint logic for True OR False."""
        or_constraint = OrConstraint(constraints=[true_constraint, false_constraint])
        result = or_constraint(time=time_array, ephemeris=mock_ephemeris, coordinate=sky_coord)
        assert np.all(result)

    def test_or_constraint_logic_false_or_false(
        self,
        mock_ephemeris: Ephemeris,
        time_array: Time,
        sky_coord: SkyCoord,
        false_constraint: ConstraintABC,
    ) -> None:
        """Test OrConstraint logic for False OR False."""
        or_constraint = OrConstraint(constraints=[false_constraint, false_constraint])
        result = or_constraint(time=time_array, ephemeris=mock_ephemeris, coordinate=sky_coord)
        assert np.all(~result)


class TestNotConstraint:
    """Test suite for NotConstraint functionality."""

    def test_not_constraint_logic_not_true(
        self,
        mock_ephemeris: Ephemeris,
        time_array: Time,
        sky_coord: SkyCoord,
        true_constraint: ConstraintABC,
    ) -> None:
        """Test NotConstraint logic for NOT True."""
        not_constraint = NotConstraint(constraint=true_constraint)
        result = not_constraint(time=time_array, ephemeris=mock_ephemeris, coordinate=sky_coord)
        assert np.all(~result)

    def test_not_constraint_logic_not_false(
        self,
        mock_ephemeris: Ephemeris,
        time_array: Time,
        sky_coord: SkyCoord,
        false_constraint: ConstraintABC,
    ) -> None:
        """Test NotConstraint logic for NOT False."""
        not_constraint = NotConstraint(constraint=false_constraint)
        result = not_constraint(time=time_array, ephemeris=mock_ephemeris, coordinate=sky_coord)
        assert np.all(result)


class TestXorConstraint:
    """Test suite for XorConstraint functionality."""

    def test_xor_constraint_empty_constraints(
        self, mock_ephemeris: Ephemeris, time_array: Time, sky_coord: SkyCoord
    ) -> None:
        """Test XorConstraint with empty constraints list."""
        xor_constraint = XorConstraint(constraints=[])
        result = xor_constraint(time=time_array, ephemeris=mock_ephemeris, coordinate=sky_coord)
        assert np.all(~result)  # All False since empty

    def test_xor_constraint_single_constraint(
        self,
        mock_ephemeris: Ephemeris,
        time_array: Time,
        sky_coord: SkyCoord,
        true_constraint: ConstraintABC,
    ) -> None:
        """Test XorConstraint with a single constraint returns same result."""
        xor_constraint = XorConstraint(constraints=[true_constraint])
        result = xor_constraint(time=time_array, ephemeris=mock_ephemeris, coordinate=sky_coord)
        expected = true_constraint(time=time_array, ephemeris=mock_ephemeris, coordinate=sky_coord)
        assert np.array_equal(result, expected)

    def test_xor_constraint_logic_true_xor_true(
        self,
        mock_ephemeris: Ephemeris,
        time_array: Time,
        sky_coord: SkyCoord,
        true_constraint: ConstraintABC,
    ) -> None:
        """Test XorConstraint logic for True XOR True."""
        xor_constraint = XorConstraint(constraints=[true_constraint, true_constraint])
        result = xor_constraint(time=time_array, ephemeris=mock_ephemeris, coordinate=sky_coord)
        assert np.all(~result)

    def test_xor_constraint_logic_true_xor_false(
        self,
        mock_ephemeris: Ephemeris,
        time_array: Time,
        sky_coord: SkyCoord,
        true_constraint: ConstraintABC,
        false_constraint: ConstraintABC,
    ) -> None:
        """Test XorConstraint logic for True XOR False."""
        xor_constraint = XorConstraint(constraints=[true_constraint, false_constraint])
        result = xor_constraint(time=time_array, ephemeris=mock_ephemeris, coordinate=sky_coord)
        assert np.all(result)

    def test_xor_constraint_logic_false_xor_true(
        self,
        mock_ephemeris: Ephemeris,
        time_array: Time,
        sky_coord: SkyCoord,
        true_constraint: ConstraintABC,
        false_constraint: ConstraintABC,
    ) -> None:
        """Test XorConstraint logic for False XOR True."""
        xor_constraint = XorConstraint(constraints=[false_constraint, true_constraint])
        result = xor_constraint(time=time_array, ephemeris=mock_ephemeris, coordinate=sky_coord)
        assert np.all(result)

    def test_xor_constraint_logic_false_xor_false(
        self,
        mock_ephemeris: Ephemeris,
        time_array: Time,
        sky_coord: SkyCoord,
        false_constraint: ConstraintABC,
    ) -> None:
        """Test XorConstraint logic for False XOR False."""
        xor_constraint = XorConstraint(constraints=[false_constraint, false_constraint])
        result = xor_constraint(time=time_array, ephemeris=mock_ephemeris, coordinate=sky_coord)
        assert np.all(~result)

    def test_xor_constraint_three_constraints_all_true(
        self,
        mock_ephemeris: Ephemeris,
        time_array: Time,
        sky_coord: SkyCoord,
        true_constraint: ConstraintABC,
    ) -> None:
        """Test XorConstraint with three True constraints."""
        xor_constraint = XorConstraint(constraints=[true_constraint, true_constraint, true_constraint])
        result = xor_constraint(time=time_array, ephemeris=mock_ephemeris, coordinate=sky_coord)
        assert np.all(result)

    def test_xor_constraint_three_constraints_two_true_one_false(
        self,
        mock_ephemeris: Ephemeris,
        time_array: Time,
        sky_coord: SkyCoord,
        true_constraint: ConstraintABC,
        false_constraint: ConstraintABC,
    ) -> None:
        """Test XorConstraint with two True and one False constraint."""
        xor_constraint = XorConstraint(constraints=[true_constraint, true_constraint, false_constraint])
        result = xor_constraint(time=time_array, ephemeris=mock_ephemeris, coordinate=sky_coord)
        assert np.all(~result)

    def test_xor_constraint_three_constraints_one_true_two_false(
        self,
        mock_ephemeris: Ephemeris,
        time_array: Time,
        sky_coord: SkyCoord,
        true_constraint: ConstraintABC,
        false_constraint: ConstraintABC,
    ) -> None:
        """Test XorConstraint with one True and two False constraints."""
        xor_constraint = XorConstraint(constraints=[true_constraint, false_constraint, false_constraint])
        result = xor_constraint(time=time_array, ephemeris=mock_ephemeris, coordinate=sky_coord)
        assert np.all(result)

    def test_xor_constraint_three_constraints_all_false(
        self,
        mock_ephemeris: Ephemeris,
        time_array: Time,
        sky_coord: SkyCoord,
        false_constraint: ConstraintABC,
    ) -> None:
        """Test XorConstraint with three False constraints."""
        xor_constraint = XorConstraint(constraints=[false_constraint, false_constraint, false_constraint])
        result = xor_constraint(time=time_array, ephemeris=mock_ephemeris, coordinate=sky_coord)
        assert np.all(~result)


class TestComplexConstraintCombinations:
    """Test suite for complex constraint combinations."""

    def test_nested_and_or_is_or_constraint(
        self,
        logical_sun_constraint: ConstraintABC,
        logical_moon_constraint: ConstraintABC,
    ) -> None:
        """Test nested (A & B) | C is an OrConstraint."""
        combined = (logical_sun_constraint & logical_moon_constraint) | logical_sun_constraint
        assert isinstance(combined, OrConstraint)

    def test_nested_and_or_first_child_is_and_constraint(
        self,
        logical_sun_constraint: ConstraintABC,
        logical_moon_constraint: ConstraintABC,
    ) -> None:
        """Test nested (A & B) | C keeps AndConstraint as first child."""
        combined = cast(
            OrConstraint,
            (logical_sun_constraint & logical_moon_constraint) | logical_sun_constraint,
        )
        assert isinstance(combined.constraints[0], AndConstraint)

    def test_nested_or_and_is_and_constraint(
        self,
        logical_sun_constraint: ConstraintABC,
        logical_moon_constraint: ConstraintABC,
    ) -> None:
        """Test nested (A | B) & C is an AndConstraint."""
        combined = (logical_sun_constraint | logical_moon_constraint) & logical_sun_constraint
        assert isinstance(combined, AndConstraint)

    def test_nested_or_and_first_child_is_or_constraint(
        self,
        logical_sun_constraint: ConstraintABC,
        logical_moon_constraint: ConstraintABC,
    ) -> None:
        """Test nested (A | B) & C keeps OrConstraint as first child."""
        combined = cast(
            AndConstraint,
            (logical_sun_constraint | logical_moon_constraint) & logical_sun_constraint,
        )
        assert isinstance(combined.constraints[0], OrConstraint)

    def test_double_negation_is_not_constraint(self, logical_sun_constraint: ConstraintABC) -> None:
        """Test double negation remains a NotConstraint wrapper."""
        double_negated = ~(~logical_sun_constraint)
        assert isinstance(double_negated, NotConstraint)

    def test_double_negation_inner_constraint_is_not_constraint(
        self, logical_sun_constraint: ConstraintABC
    ) -> None:
        """Test double negation inner constraint is also NotConstraint."""
        double_negated = cast(NotConstraint, ~(~logical_sun_constraint))
        assert isinstance(double_negated.constraint, NotConstraint)

    def test_complex_and_or_combination(
        self,
        mock_ephemeris: Ephemeris,
        time_array: Time,
        sky_coord: SkyCoord,
        true_constraint: ConstraintABC,
        false_constraint: ConstraintABC,
    ) -> None:
        """Test complex combination: (A & B) | C."""
        # (True & True) | False = True | False = True
        combined = (true_constraint & true_constraint) | false_constraint
        result = combined(time=time_array, ephemeris=mock_ephemeris, coordinate=sky_coord)
        assert np.all(result)

    def test_complex_not_combination(
        self, mock_ephemeris: Ephemeris, time_array: Time, sky_coord: SkyCoord, true_constraint: ConstraintABC
    ) -> None:
        """Test complex combination: ~(A & B)."""
        # ~(True & True) = ~True = False
        combined = ~(true_constraint & true_constraint)
        result = combined(time=time_array, ephemeris=mock_ephemeris, coordinate=sky_coord)
        assert np.all(~result)

    def test_demorgan_law(
        self,
        mock_ephemeris: Ephemeris,
        time_array: Time,
        sky_coord: SkyCoord,
        true_constraint: ConstraintABC,
        false_constraint: ConstraintABC,
    ) -> None:
        """Test De Morgan's law: ~(A & B) == ~A | ~B."""
        constraint_a = true_constraint
        constraint_b = false_constraint

        # ~(A & B)
        left = ~(constraint_a & constraint_b)
        result_left = left(time=time_array, ephemeris=mock_ephemeris, coordinate=sky_coord)

        # ~A | ~B
        right = ~constraint_a | ~constraint_b
        result_right = right(time=time_array, ephemeris=mock_ephemeris, coordinate=sky_coord)

        assert np.array_equal(result_left, result_right)

    def test_nested_xor_combination_is_or_constraint(
        self,
        logical_sun_constraint: ConstraintABC,
        logical_moon_constraint: ConstraintABC,
    ) -> None:
        """Test nested (A ^ B) | C is an OrConstraint."""
        combined = (logical_sun_constraint ^ logical_moon_constraint) | logical_sun_constraint
        assert isinstance(combined, OrConstraint)

    def test_nested_xor_combination_first_child_is_xor_constraint(
        self,
        logical_sun_constraint: ConstraintABC,
        logical_moon_constraint: ConstraintABC,
    ) -> None:
        """Test nested (A ^ B) | C keeps XorConstraint as first child."""
        combined = cast(
            OrConstraint,
            (logical_sun_constraint ^ logical_moon_constraint) | logical_sun_constraint,
        )
        assert isinstance(combined.constraints[0], XorConstraint)

    def test_complex_xor_combination_true_xor_false_and_true(
        self,
        mock_ephemeris: Ephemeris,
        time_array: Time,
        sky_coord: SkyCoord,
        true_constraint: ConstraintABC,
        false_constraint: ConstraintABC,
    ) -> None:
        """Test (True ^ False) & True evaluates to True."""
        combined = (true_constraint ^ false_constraint) & true_constraint
        result = combined(time=time_array, ephemeris=mock_ephemeris, coordinate=sky_coord)
        assert np.all(result)

    def test_complex_xor_combination_true_xor_true_and_true(
        self,
        mock_ephemeris: Ephemeris,
        time_array: Time,
        sky_coord: SkyCoord,
        true_constraint: ConstraintABC,
    ) -> None:
        """Test (True ^ True) & True evaluates to False."""
        combined = (true_constraint ^ true_constraint) & true_constraint
        result = combined(time=time_array, ephemeris=mock_ephemeris, coordinate=sky_coord)
        assert np.all(~result)



from datetime import datetime

import astropy.units as u  # type: ignore[import-untyped]
import pytest
from astropy.coordinates import AltAz, SkyCoord  # type: ignore[import-untyped]

from across.tools.ephemeris import Ephemeris


@pytest.fixture
def az_zero_alt_forty_five_sky_coord(ground_ephemeris: Ephemeris, ephemeris_begin: datetime) -> SkyCoord:
    """Fixture for a sky coordinate at 45 deg altitude and 50 deg azimuth."""
    return SkyCoord(
        AltAz(
            alt=45 * u.deg,
            az=50 * u.deg,
            location=ground_ephemeris.earth_location,
            obstime=ephemeris_begin,
        )
    )


@pytest.fixture
def az_eight_alt_five_sky_coord(ground_ephemeris: Ephemeris, ephemeris_begin: datetime) -> SkyCoord:
    """Fixture for a sky coordinate at 8 deg altitude and 5 deg azimuth."""
    return SkyCoord(
        AltAz(
            alt=8 * u.deg,
            az=5 * u.deg,
            location=ground_ephemeris.earth_location,
            obstime=ephemeris_begin,
        )
    )



from typing import Literal, cast

import numpy as np
import pytest
from astropy.coordinates import SkyCoord  # type: ignore[import-untyped]
from astropy.time import Time  # type: ignore[import-untyped]

from across.tools.core.enums.constraint_type import ConstraintType
from across.tools.ephemeris import Ephemeris
from across.tools.visibility.constraints import (
    AndConstraint,
    MoonAngleConstraint,
    NotConstraint,
    OrConstraint,
    SunAngleConstraint,
    XorConstraint,
)
from across.tools.visibility.constraints.base import ConstraintABC


class DummyConstraint(ConstraintABC):
    """Dummy constraint for testing purposes."""

    short_name: str = "Dummy"
    name: Literal[ConstraintType.UNKNOWN] = ConstraintType.UNKNOWN
    min_angle: float | None = None
    max_angle: float | None = None

    def __call__(self, time: Time, ephemeris: Ephemeris, coordinate: SkyCoord) -> np.typing.NDArray[np.bool_]:
        """Dummy implementation of the constraint."""
        return np.zeros(len(time), dtype=bool)


class TrueConstraint(DummyConstraint):
    """Constraint that always returns True (always violated)."""

    short_name: Literal["True"] = "True"

    def __call__(self, time: Time, ephemeris: Ephemeris, coordinate: SkyCoord) -> np.typing.NDArray[np.bool_]:
        """Always return True."""
        return np.ones(len(time), dtype=bool)


class FalseConstraint(DummyConstraint):
    """Constraint that always returns False (never violated)."""

    short_name: Literal["False"] = "False"

    def __call__(self, time: Time, ephemeris: Ephemeris, coordinate: SkyCoord) -> np.typing.NDArray[np.bool_]:
        """Always return False."""
        return np.zeros(len(time), dtype=bool)


@pytest.fixture
def true_constraint() -> TrueConstraint:
    """Fixture for a TrueConstraint instance that always returns True."""
    return TrueConstraint()


@pytest.fixture
def false_constraint() -> FalseConstraint:
    """Fixture for a FalseConstraint instance that always returns False."""
    return FalseConstraint()


@pytest.fixture
def logical_sun_constraint() -> SunAngleConstraint:
    """Sun constraint used by logical-operator structure tests."""
    return SunAngleConstraint(min_angle=45)


@pytest.fixture
def logical_moon_constraint() -> MoonAngleConstraint:
    """Moon constraint used by logical-operator structure tests."""
    return MoonAngleConstraint(min_angle=10)


@pytest.fixture
def logical_and_constraint(
    logical_sun_constraint: SunAngleConstraint,
    logical_moon_constraint: MoonAngleConstraint,
) -> AndConstraint:
    """(Sun & Moon) combined constraint."""
    return cast(AndConstraint, logical_sun_constraint & logical_moon_constraint)


@pytest.fixture
def logical_or_constraint(
    logical_sun_constraint: SunAngleConstraint,
    logical_moon_constraint: MoonAngleConstraint,
) -> OrConstraint:
    """(Sun | Moon) combined constraint."""
    return cast(OrConstraint, logical_sun_constraint | logical_moon_constraint)


@pytest.fixture
def logical_xor_constraint(
    logical_sun_constraint: SunAngleConstraint,
    logical_moon_constraint: MoonAngleConstraint,
) -> XorConstraint:
    """(Sun ^ Moon) combined constraint."""
    return cast(XorConstraint, logical_sun_constraint ^ logical_moon_constraint)


@pytest.fixture
def logical_not_constraint(logical_sun_constraint: SunAngleConstraint) -> NotConstraint:
    """(~Sun) negated constraint."""
    return cast(NotConstraint, ~logical_sun_constraint)



# Copyright © 2023 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.


import logging
import re
from datetime import datetime, timedelta
from typing import TypedDict

from httpx import HTTPStatusError
from spacetrack import AuthenticationError, SpaceTrackClient  # type: ignore[import-untyped]

from ..core.config import config
from ..core.schemas.tle import TLE
from .exceptions import SpaceTrackAuthenticationError


class NoradSatellite(TypedDict):
    """Satellite specification with name and NORAD ID."""

    name: str
    id: int


class TLEFetch:
    """
    Fetches Two-Line Element (TLE) data for one or more satellites at a specific epoch.
    Requires a Space-Track.org account to access the TLE data. If no spacetrack
    user or password is provided, the class will attempt to use the environment
    variables SPACETRACK_USER and SPACETRACK_PWD.

    Parameters
    ----------
    satellites : list[NoradSatellite]
        List of satellites to query, each with 'name' and 'id' keys.
        Only the id is used to query against spacetrack_client.gp norad_cat_id
    epoch : datetime
        Epoch of TLE to retrieve.
    spacetrack_user : str, optional
        Space-Track.org username. Falls back to SPACETRACK_USER config.
    spacetrack_pwd : str, optional
        Space-Track.org password. Falls back to SPACETRACK_PWD config.
    spacetrack_base_url : str, optional
        Optional Space-Track API base URL override. Default is None.

    Attributes
    ----------
    satellites : list[NoradSatellite]
        List of satellites with name and NORAD ID.
    epoch : datetime
        Epoch of the TLE
    spacetrack_user : str, optional
        Space-Track.org username
    spacetrack_pwd  : str, optional
        Space-Track.org password
    spacetrack_base_url : str, optional
        Space-Track API base URL override

    Methods
    -------
    get
        Get TLEs for given epoch

    Raises
    ------
    TypeError
        - satellites must be a list of dicts with 'name' and 'id' keys
        - each satellite must be a dict
        - each satellite dict must have 'name' and 'id' keys
        - satellite 'name' must be a string
        - satellite 'id' must be an integer
    ValueError
        - satellites list cannot be empty
    """

    # Configuration parameters
    satellites: list[NoradSatellite]
    epoch: datetime
    spacetrack_user: str | None
    spacetrack_pwd: str | None
    spacetrack_base_url: str | None

    def __init__(
        self,
        satellites: list[NoradSatellite],
        epoch: datetime,
        spacetrack_user: str | None = None,
        spacetrack_pwd: str | None = None,
        spacetrack_base_url: str | None = None,
    ):
        if not isinstance(satellites, list):
            raise TypeError("satellites must be a list of dicts with 'name' and 'id' keys")
        if not satellites:
            raise ValueError("satellites list cannot be empty")
        for sat in satellites:
            if not isinstance(sat, dict):
                raise TypeError("each satellite must be a dict")
            if "name" not in sat or "id" not in sat:
                raise TypeError("each satellite dict must have 'name' and 'id' keys")
            if not isinstance(sat["name"], str):
                raise TypeError("satellite 'name' must be a string")
            if not isinstance(sat["id"], int):
                raise TypeError("satellite 'id' must be an integer")

        self.satellites = satellites
        self.epoch = epoch
        self.spacetrack_user = spacetrack_user or config.SPACETRACK_USER
        self.spacetrack_pwd = spacetrack_pwd or config.SPACETRACK_PWD
        self.spacetrack_base_url = spacetrack_base_url

    @staticmethod
    def _extract_norad_id_from_tle1(tle1: str) -> int | None:
        """Extract NORAD ID from the first segment of TLE line 1."""
        match = re.match(r"^1\s+(\d+)", tle1)
        return int(match.group(1)) if match else None

    def get(self) -> list[TLE]:
        """
        Return TLE data for requested NORAD ID(s), within +/- 7 days.

        Aggregates response by NORAD ID
        by scanning the first section of TLE line 1, and only the
        newest TLE for each requested NORAD ID is retained.

        Returns
        -------
        list[TLE]
            TLE objects matching the requested NORAD ID(s). Empty when no data
            is found.

        Raises
        ------
        SpaceTrackAuthenticationError
            If space-track.org authentication fails.
        """
        if not self.satellites:
            return []

        # Build space-track.org query
        epoch_start = self.epoch - timedelta(days=7)
        epoch_stop = self.epoch + timedelta(days=7)

        spacetrack_client_args = {"identity": self.spacetrack_user, "password": self.spacetrack_pwd}

        # set base_url when provided upstream to override space-track server to access
        if self.spacetrack_base_url is not None:
            spacetrack_client_args["base_url"] = self.spacetrack_base_url

        # Log into space-track.org
        with SpaceTrackClient(**spacetrack_client_args) as spacetrack_client:
            try:
                spacetrack_client.authenticate()
            except (AuthenticationError, HTTPStatusError) as e:
                raise SpaceTrackAuthenticationError("space-track.org authentication failed.") from e

            # Extract NORAD IDs for the query
            norad_ids = [sat["id"] for sat in self.satellites]

            # Fetch the TLEs between the requested epochs
            tletext = spacetrack_client.gp(
                norad_cat_id=norad_ids,
                orderby="epoch desc",
                format="tle",
                epoch=f">{epoch_start},<{epoch_stop}",
            )

        # Check if we got a return
        if tletext == "":
            return []

        # Split the TLEs into individual lines and pair line1/line2 entries.
        lines = tletext.splitlines()
        tle_pairs = [(lines[i], lines[i + 1]) for i in range(0, len(lines), 2)]

        # Build map from norad_id to satellite name
        sat_name_map = {sat["id"]: sat["name"] for sat in self.satellites}
        requested_ids = set(sat_name_map.keys())
        seen_ids: set[int] = set()
        tles: list[TLE] = []

        # Filter to the first occurrence of each requested NORAD ID (newest due to orderby parameter)
        for tle1, tle2 in tle_pairs:
            parsed_id = self._extract_norad_id_from_tle1(tle1)
            if parsed_id is None:
                logging.warning("Could not parse norad_id from space-track", {"tle1": tle1, "tle2": tle2})
                continue
            if parsed_id not in requested_ids:
                logging.warning(f"Received norad_id {parsed_id} TLE from space-track that was not requested")
                continue
            if parsed_id not in seen_ids:
                tles.append(
                    TLE(
                        satellite_name=sat_name_map.get(parsed_id),
                        norad_id=parsed_id,
                        tle1=tle1,
                        tle2=tle2,
                    )
                )

                seen_ids.add(parsed_id)

                # all requested satellite ids have been parsed into TLEs from the response
                if seen_ids == requested_ids:
                    break

        return tles


def get_tle(
    satellites: list[NoradSatellite],
    epoch: datetime,
    spacetrack_user: str | None = None,
    spacetrack_pwd: str | None = None,
    spacetrack_base_url: str | None = None,
) -> list[TLE]:
    """
    Gets the Two-Line Element (TLE) data for one or more satellites at a specific epoch.
    Credentials for space-track.org can be provided as arguments, or they can
    be set as environment variables SPACETRACK_USER and SPACETRACK_PWD.

    Parameters
    ----------
    satellites : list[NoradSatellite]
        List of satellites to query, each with 'name' and 'id' keys.
    epoch : datetime
        The epoch timestamp for which to retrieve the TLE data.
    spacetrack_user : str, optional
        space-Track.org username.
    spacetrack_pwd : str, optional
        space-Track.org password.
    spacetrack_base_url : str, optional
        Optional Space-Track API base URL override.

    Returns
    -------
    list[TLE]
        TLE objects matching the requested satellites. Empty if no data is
        found.

    Raises
    ------
    TypeError
        - satellites must be a list of dicts with 'name' and 'id' keys
        - each satellite must be a dict
        - each satellite dict must have 'name' and 'id' keys
        - satellite 'name' must be a string
        - satellite 'id' must be an integer
    ValueError
        - satellites list cannot be empty
    """

    tle = TLEFetch(
        satellites=satellites,
        epoch=epoch,
        spacetrack_user=spacetrack_user,
        spacetrack_pwd=spacetrack_pwd,
        spacetrack_base_url=spacetrack_base_url,
    )
    return tle.get()



from collections import Counter

import numpy as np
import numpy.typing as npt


def find_duplicates(input_list: list[int]) -> list[int]:
    """
    Finds duplicates in a list.
    Taken from https://stackoverflow.com/questions/9835762/
    """
    return [item for item, count in Counter(input_list).items() if count > 1]


def x_rot(theta_deg: float) -> npt.NDArray[np.float64]:
    """
    Performs matrix rotation around the cartesian x direction
    """
    theta = np.deg2rad(theta_deg)
    return np.asarray([[1, 0, 0], [0, np.cos(theta), -np.sin(theta)], [0, np.sin(theta), np.cos(theta)]])


def y_rot(theta_deg: float) -> npt.NDArray[np.float64]:
    """
    Performs matrix rotation around the cartesian y direction
    """
    theta = np.deg2rad(theta_deg)
    return np.asarray([[np.cos(theta), 0, np.sin(theta)], [0, 1, 0], [-np.sin(theta), 0, np.cos(theta)]])


def z_rot(theta_deg: float) -> npt.NDArray[np.float64]:
    """
    Performs matrix rotation around the cartesian z direction
    """
    theta = np.deg2rad(theta_deg)
    return np.asarray([[np.cos(theta), -np.sin(theta), 0], [np.sin(theta), np.cos(theta), 0], [0, 0, 1]])



from datetime import datetime, timedelta

import rust_ephem
from astropy.time import Time, TimeDelta  # type: ignore[import-untyped]

from ..core.schemas.tle import TLE
from .base import Ephemeris


class TLEEphemeris(Ephemeris):
    """
    TLE (Two-Line Element) based satellite ephemeris calculator.

    This class implements satellite position and velocity calculations using
    TLE data through the SGP4 propagator. It converts TEME (True Equator Mean
    Equinox) coordinates to ITRS (International Terrestrial Reference System)
    and GCRS (Geocentric Celestial Reference System). ITRS coordinates are
    stored as astropy EarthLocation objects, while GCRS coordinates are stored as
    astropy SkyCoord objects.

    Parameters
    ----------
    begin : Union[datetime, Time]
        Start time for ephemeris calculations
    end : Union[datetime, Time]
        End time for ephemeris calculations
    step_size : Union[int, TimeDelta, timedelta], optional
        Time step between calculations, defaults to 60 seconds
    tle : Optional[TLE], optional
        TLE data object containing orbital elements, defaults to None

    Attributes
    ----------
    tle : Optional[TLE]
        TLE data object used for calculations
    earth_location : EarthLocation
        Satellite positions in ITRS coordinates
    gcrs : SkyCoord
        Satellite positions in GCRS coordinates

    Methods
    -------
    prepare_data()
        Calculates satellite ephemeris using TLE data and SGP4 propagator

    Raises
    ------
    ValueError
        If no TLE data is provided when preparing calculations

    Notes
    -----
    The calculation process involves:
    1. Converting TLE to Satrec object
    2. Calculating TEME positions using SGP4
    3. Converting TEME to ITRS coordinates
    4. Converting ITRS to GCRS coordinates
    """

    # TLE for calculating LEO satellites
    tle: TLE | None = None

    def __init__(
        self,
        begin: datetime | Time,
        end: datetime | Time,
        step_size: int | TimeDelta | timedelta = 60,
        tle: TLE | None = None,
    ) -> None:
        super().__init__(begin, end, step_size)
        self.tle = tle

    def prepare_data(self) -> None:
        """Calculate ephemeris based on TLE data"""
        # Check if TLE is loaded
        if self.tle is None:
            raise ValueError("No TLE provided")

        # Ensure planetary ephemeris data is prepared
        rust_ephem.ensure_planetary_ephemeris()

        # Calculate ephemeris using rust-ephem library
        self._tle_ephem = rust_ephem.TLEEphemeris(
            begin=self.begin.datetime,
            end=self.end.datetime,
            step_size=int(self._step_seconds),
            tle1=self.tle.tle1,
            tle2=self.tle.tle2,
        )

        # Calculate satellite position in ITRS coordinate system
        self.earth_location = self._tle_ephem.itrs.earth_location

        # Calculate satellite position in GCRS coordinate system vector as
        # array of x,y,z vectors in units of km, and velocity vector as array
        # of x,y,z vectors in units of km/s
        self.gcrs = self._tle_ephem.gcrs

    def _calc(self) -> None:
        """
        Calculate ephemeris data based on the coordinates computed by
        prepare_data(). Overloads the base class method to as rust-ephem already
        computes all necessary data during prepare_data().
        """
        # Calculate the position of the Moon relative to the spacecraft
        self.moon = self._tle_ephem.moon

        # Calculate the position of the Sun relative to the spacecraft
        self.sun = self._tle_ephem.sun

        # Calculate the position of the Earth relative to the spacecraft
        self.earth = self._tle_ephem.earth

        # Get the longitude, latitude, height, and distance (from center of
        # Earth) of the satellite from the computed ephemeris.
        self.latitude = self._tle_ephem.latitude
        self.longitude = self._tle_ephem.longitude
        self.height = self._tle_ephem.height
        self.distance = self.gcrs.distance

        # Calculate Earth's angular radius from observatory, capped at 90 degrees
        self.earth_radius_angle = self._tle_ephem.earth_radius

        # Similarly calculate the angular radii of the Sun and the Moon, capped at 90 degrees
        self.moon_radius_angle = self._tle_ephem.moon_radius
        self.sun_radius_angle = self._tle_ephem.sun_radius


def compute_tle_ephemeris(
    begin: datetime | Time,
    end: datetime | Time,
    step_size: int | timedelta | TimeDelta,
    tle: TLE,
) -> Ephemeris:
    """
    Compute the ephemeris for a space object using TLE data.

    Parameters
    ----------
    begin : Union[datetime, Time]
        The start time for the ephemeris computation.
    end : Union[datetime, Time]
        The end time for the ephemeris computation.
    step_size : Union[int, timedelta, TimeDelta]
        Time step size between ephemeris points, in seconds if int
    tle : TLE
        The TLE data entry for the space object.

    Returns
    -------
    Ephemeris
        The computed ephemeris object containing the position and velocity data.
    """
    # Compute the ephemeris using the TLEEphemeris class
    ephemeris = TLEEphemeris(tle=tle, begin=begin, end=end, step_size=step_size)
    ephemeris.compute()
    return ephemeris



from collections.abc import Generator
from unittest.mock import MagicMock

import astropy.units as u  # type: ignore[import-untyped]
import pytest
from astropy.coordinates import SkyCoord  # type: ignore[import-untyped]
from astropy.time import Time  # type: ignore[import-untyped]

from across.tools.visibility.constraints.bright_star import BrightStarConstraint


@pytest.fixture
def mock_get_bright_stars(
    mock_bright_stars: list[tuple[SkyCoord, float]],
    monkeypatch: pytest.MonkeyPatch,
) -> Generator[MagicMock, None, None]:
    """Patch get_bright_stars to prevent internet access during tests."""
    mock = MagicMock(return_value=mock_bright_stars)
    monkeypatch.setattr("across.tools.visibility.constraints.bright_star.get_bright_stars", mock)
    yield mock


@pytest.fixture
def bright_star_constraint() -> BrightStarConstraint:
    """Default bright star constraint used in call tests."""
    return BrightStarConstraint(min_separation=5.0)


@pytest.fixture
def sirius_coord() -> SkyCoord:
    """Coordinate at Sirius position."""
    return SkyCoord(ra="06h45m08.9s", dec="-16d42m58.0s")


@pytest.fixture
def near_sirius_coord(sirius_coord: SkyCoord) -> SkyCoord:
    """Coordinate near Sirius (within default threshold)."""
    return SkyCoord(ra=sirius_coord.ra + 1 * u.deg, dec=sirius_coord.dec)


@pytest.fixture
def far_from_bright_stars_coord() -> SkyCoord:
    """Coordinate far from bright stars."""
    return SkyCoord(ra=0 * u.deg, dec=80 * u.deg)


@pytest.fixture
def mixed_bright_star_coords() -> SkyCoord:
    """Coordinates containing one point at Sirius."""
    return SkyCoord(
        ra=["00h00m00s", "06h45m08.9s", "10h00m00s"],
        dec=["00d00m00s", "-16d42m58.0s", "00d00m00s"],
    )


@pytest.fixture
def mixed_bright_star_times() -> Time:
    """Timestamps aligned with mixed bright star coordinates."""
    return Time(["2024-01-01T00:00:00", "2024-01-01T01:00:00", "2024-01-01T02:00:00"])



from typing import Literal

import astropy.units as u  # type: ignore[import-untyped]
import numpy as np
import numpy.typing as npt
from astropy.coordinates import SkyCoord  # type: ignore[import-untyped]
from astropy.time import Time  # type: ignore[import-untyped]
from pydantic import Field

from ...core.enums import ConstraintType
from ...ephemeris import Ephemeris
from .base import ConstraintABC


class GalacticBulgeConstraint(ConstraintABC):
    """
    Constraint that avoids observing too close to the Galactic Bulge.

    The Galactic Bulge is a crowded region with extremely high stellar density,
    making it difficult to observe individual sources. This constraint ensures
    observations avoid this region.

    Parameters
    ----------
    min_separation : float
        Minimum angular separation (degrees) required from the Galactic Bulge.
        Observations closer than this will be constrained.

    Methods
    -------
    __call__(time, ephemeris, coordinate)
        Checks if the coordinate is too close to the Galactic Bulge.
    """

    name: Literal[ConstraintType.GALACTIC_BULGE] = ConstraintType.GALACTIC_BULGE
    short_name: Literal["Galactic Bulge"] = "Galactic Bulge"
    min_separation: float = Field(
        default=10.0, gt=0, le=180, description="Minimum angular separation (degrees) from Galactic Bulge"
    )

    def __call__(self, time: Time, ephemeris: Ephemeris, coordinate: SkyCoord) -> npt.NDArray[np.bool_]:
        """
        Check if the coordinate is too close to the Galactic Bulge.

        Parameters
        ----------
        time : Time
            The time (not used for galactic coordinates).
        ephemeris : Ephemeris
            The ephemeris (not used for galactic coordinates).
        coordinate : SkyCoord
            The coordinate to check.

        Returns
        -------
        npt.NDArray[np.bool_]
            Boolean array where True indicates the coordinate violates the constraint
            (is too close to the Galactic Bulge).
        """
        # Galactic Bulge coordinates (J2000)
        galactic_bulge = SkyCoord(ra="17h45m40.04s", dec="-29d00m28.1s", frame="icrs")

        # Calculate angular separation
        separation = coordinate.separation(galactic_bulge)

        # Record the separation in computed values
        self.computed_values.galactic_bulge_separation = separation

        # Constrain observations closer than the minimum separation
        in_constraint: npt.NDArray[np.bool_] = separation < (self.min_separation * u.deg)

        return in_constraint



from datetime import datetime, timedelta
from typing import Any

from pydantic import Field, model_validator

from .base import BaseSchema


class TLEBase(BaseSchema):
    """
    A base schema representing a Two-Line Element (TLE) set for satellite tracking.
    Parameters
    ----------
    norad_id : int or None
        The NORAD Catalog Number (SATCAT) that uniquely identifies the satellite.
    satellite_name : str or None
        The name or designation of the satellite.
    tle1 : str
        The first line of the TLE set, must be exactly 69 characters long.
        Contains information about satellite epoch, decay rate, etc.
    tle2 : str
        The second line of the TLE set, must be exactly 69 characters long.
        Contains orbital elements like inclination, eccentricity, etc.
    Notes
    -----
    TLE is a data format encoding a set of orbital elements for Earth-orbiting objects.
    Each TLE line must be exactly 69 characters long per NORAD specification.
    """

    norad_id: int | None = None
    satellite_name: str | None = None
    tle1: str = Field(min_length=69, max_length=69, pattern=r"1 .{67}")
    tle2: str = Field(min_length=69, max_length=69, pattern=r"2 .{67}")


class TLE(TLEBase):
    """
    Two Line Element (TLE) data representation.

    This class represents a Two Line Element Set, which is a data format used to convey
    sets of orbital elements that describe the orbits of Earth-orbiting satellites.

    Parameters
    ----------
    norad_id : int
        The NORAD Catalog Number (SATCAT) that uniquely identifies the satellite
    satellite_name : str
        The name of the satellite (used as Partition Key)
    tle1 : str
        First line of the TLE, must be exactly 69 characters
    tle2 : str
        Second line of the TLE, must be exactly 69 characters
    epoch : datetime
        The epoch timestamp calculated from the TLE data

    Notes
    -----
    TLE format specifications can be found at:
    https://celestrak.org/NORAD/documentation/tle-fmt.php

    Examples
    --------
    >>> tle = TLE(
    ...     satname="ISS (ZARYA)",
    ...     tle1="1 25544U 98067A   08264.51782528 -.00002182  00000-0 -11606-4 0  2927",
    ...     tle2="2 25544  51.6416 247.4627 0006703 130.5360 325.0288 15.72125391563537"
    ... )
    """

    epoch: datetime = datetime(2000, 1, 1)

    @model_validator(mode="before")
    @classmethod
    def validate_tle(cls, values: dict[str, Any]) -> dict[str, Any]:
        """
        Validate the TLE data.
        This checks whether the TLE lines are correctly formatted and contain
        valid information, as well as calculate the epoch.
        Returns
        -------
            The calculated epoch of the TLE.
        """
        # Extract year and days from TLE
        tleepoch = values["tle1"].split()[3]
        tleyear = int(tleepoch[:2])
        days = float(tleepoch[2:]) - 1

        # Calculate epoch date
        year = 2000 + tleyear if tleyear < 57 else 1900 + tleyear
        tle_epoch = datetime(year, 1, 1) + timedelta(days=days)

        # If epoch isn't set in input, set it
        if not values.get("epoch"):
            values["epoch"] = tle_epoch
        else:
            # Check that epoch given matches one derived from TLE
            if values["epoch"] != tle_epoch:
                raise ValueError("Epoch derived from TLE does not match given epoch.")

        return values



"""
NASA Docket No. GSC-19,469-1, and identified as "Astrophysics Cross-Observatory
Science Support (ACROSS) System

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software distributed
under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.“

The copyright notice to be included in the software is as follows:

Copyright © 2025 United States Government as represented by the Administrator
of the National Aeronautics and Space Administration and The Penn State
Research Foundation.  All rights reserved. This software is licensed under the
Apache 2.0 License.
"""

__path__ = __import__("pkgutil").extend_path(__path__, __name__)
